// Copyright (C) 1996 by Florian Raemy & LCAV, Swiss Federal Institute of
// Technology, Lausanne, Switzerland.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//
// Visit Florian's Home Page http://studwww.eurecom.fr/~raemy and
// LCAV Department of EPFL: http://lcavwww.epfl.ch

import java.lang.*;
import java.awt.*;
import java.net.*;
import java.io.*;
import java.util.*;
import java.awt.image.*;

/**
 * Graphix - A collection of graphics functions
 *
 * Feature list:
 * Array returns (R,G,B).
 * E-Z Image Producing.
 * RGB to YUV colorspace conversion.
 * YUV to RGB colorspace conversion.
 *
 */

public class Grafixs
{
    public Image imageobj;
    public int imageHeight;
    public int imageWidth;

    /**
     * Gives GraphixTools a preloaded image object.
     * The constructor calculates all the image parameters from this
     * passed image. (Dimensions, size)
     */

    public Grafixs(Image image)
    {
        imageobj = image;
        imageWidth = image.getWidth(null);
        imageHeight = image.getHeight(null);
    }

    /**
     * This method returns an array[][] of 8-bit red values. The dimensions are
     * the array bounds ex. (320 x 200).
     * @returns An array[][] of red pixel values
     */
    public int[][] getRedArray()
    {
        int values[] = new int[imageWidth * imageHeight];
    	PixelGrabber grabber = new PixelGrabber(imageobj.getSource(), 0, 0, imageWidth, imageHeight, values, 0, imageWidth);

    	try
    	{
    	    if(grabber.grabPixels() != true)
    	    {
    		    try
    		    {
    		        throw new AWTException("Grabber returned false: " + grabber.status());
    		    }
    		    catch (Exception e) {};
    		}
    	}
    	catch (InterruptedException e) {};
    	int r[][] = new int[imageWidth][imageHeight];
  		int index = 0;
    	for (int y = 0; y < imageHeight; ++y)
    	{
    	    for (int x = 0; x < imageWidth; ++x)
    	    {
        	    r[x][y] = ((values[index] & 0x00ff0000) >> 16);
        		index++;
    	    }
    	}

    	return r;
    }

    /**
     * This method returns an array[][] of 8-bit green values. The dimensions are
     * the array bounds ex. (320 x 200).
     * @returns An array[][] of green pixel values.
     */
    public int[][] getGreenArray()
    {
        int values[] = new int[imageWidth * imageHeight];
    	PixelGrabber grabber = new PixelGrabber(imageobj.getSource(), 0, 0, imageWidth, imageHeight, values, 0, imageWidth);
    	try
    	{
    	    if(grabber.grabPixels() != true)
    	    {
    	        try
    	        {
    		        throw new AWTException("Grabber returned false: " + grabber.status());
    		    }
    		    catch (Exception e) {};
    		}
    	}
    	catch (InterruptedException e) {};
    	int g[][] = new int[imageWidth][imageHeight];
  		int index = 0;
    	for (int y = 0; y < imageHeight; ++y)
    	{
    	    for (int x = 0; x < imageWidth; ++x)
    	    {
        		g[x][y] = ((values[index] & 0x0000ff00) >> 8);
        		++index;
    	    }
    	}

    	return g;
    }

    /**
     * This method returns an array[][] of 8-bit blue values. The dimensions are
     * the array bounds, ex. (320 x 200).
     * @returns An array[][] of blue pixel values.
     */
    public int[][] getBlueArray()
    {
        int values[] = new int[imageWidth * imageHeight];
    	PixelGrabber grabber = new PixelGrabber(imageobj.getSource(), 0, 0, imageWidth, imageHeight, values, 0, imageWidth);

    	try
    	{
    	    if(grabber.grabPixels() != true)
    	    {
    	        try
    	        {
    		        throw new AWTException("Grabber returned false: " + grabber.status());
    		    }
    		    catch (Exception e) {};
    		}
    	}
    	catch (InterruptedException e) {};

    	int b[][] = new int[imageWidth][imageHeight];

  		int index = 0;
    	for (int y = 0; y < imageHeight; ++y)
    	{
    	    for (int x = 0; x < imageWidth; ++x)
    	    {
        		b[x][y] = (values[index] & 0x000000ff);

        		++index;
    	    }
    	}

    	return b;
    }

    /**
     * This method converts an RGB colorspace to a YUV colorspace. (YUV is more
     * technically correct when referred to a YCbCr colorspace.)
     * @param redArray An array of 0..255 red pixel values.
     * @param greenArray An array of 0..255 green pixel values.
     * @param blueArray An array of 0..255 blue pixel values.
     * @returns Y The Y component
     */

     public float[][] convertRGBtoY(int[][] redArray, int[][] greenArray, int[][] blueArray)
    {
        int i;
        int j;
        float Y[][] = new float[imageWidth][imageHeight];

        for (i=0; i<imageWidth; i++)
        {
            for (j=0; j<imageHeight; j++)
            {

                Y[i][j] = 16 + (float)(0.8588*(0.299 * (float)redArray[i][j] + 0.587 * (float)greenArray[i][j] + 0.114 * (float)blueArray[i][j] ));

            }
        }

        return Y;
    }

     /**
     * This method converts an RGB colorspace to a YUV colorspace. (YUV is more
     * technically correct when referred to a YCbCr colorspace.)
     * @param redArray An array of 0..255 red pixel values.
     * @param greenArray An array of 0..255 green pixel values.
     * @param blueArray An array of 0..255 blue pixel values.
     * @returns Cb The Cb component.
     */

     public float[][] convertRGBtoCb(int[][] redArray, int[][] greenArray, int[][] blueArray)
    {
        int i;
        int j;
        float Cb[][] = new float[imageWidth][imageHeight];

        for (i=0; i<imageWidth; i++)
        {
            for (j=0; j<imageHeight; j++)
            {


                Cb[i][j] =128 + (float)(0.8784*(-0.1687 * (float)redArray[i][j] - 0.3313 * (float)greenArray[i][j] + 0.5 * (float)blueArray[i][j]));
            }
        }

        return Cb;
    }

    /**
     * This method converts an RGB colorspace to a YUV colorspace. (YUV is more
     * technically correct when referred to a YCbCr colorspace.)
     * @param redArray An array of 0..255 red pixel values.
     * @param greenArray An array of 0..255 green pixel values.
     * @param blueArray An array of 0..255 blue pixel values.
     * @returns Cr The Cr component.
     */


     public float[][] convertRGBtoCr(int[][] redArray, int[][] greenArray, int[][] blueArray)
    {
        int i;
        int j;
        float Cr[][] = new float[imageWidth][imageHeight];


        for (i=0; i<imageWidth; i++)
        {
            for (j=0; j<imageHeight; j++)
            {
                   Cr[i][j] = 128 + (float)(0.8784*(0.5 * (float)redArray[i][j] - 0.4187 * (float)greenArray[i][j] - 0.0813 * (float)blueArray[i][j]));
            }
        }

        return Cr;
    }


     /**
     * This method converts a YUV (aka YCbCr) colorspace to a RGB colorspace.
     * This is handy when trying to reconstruct an image in Java from YCbCr transmitted
     * data. This routine expects the data to fall in the standard PC 0..255 range
     * per pixel, with the array dimensions corresponding to the imageWidth and imageHeight.
     * These variables are either set manually in the case of a null constructor,
     * or they are automatically calculated from the image parameter constructor.
     * @param Y The Y component set.
     * @param Cb The Cb component set.
     * @param Cr The Cr component set.
     * @returns R The R component.
     */
    public int[][] convertYCbCrtoR(int[][] Y, int[][] Cb, int[][] Cr)
    {
        int i;
        int j;
        int R[][] = new int[imageWidth][imageHeight];
        System.out.println(" YCbCr to R avec largeur="+imageWidth+" et hauteur"+imageHeight);
        for (i=0; i<imageWidth; i++)
        {

            for (j=0; j<imageHeight; j++)
            {
                R[i][j] = (int)( ((1.164*((float)Y[i][j]-16.0))) + (1.596*((float)Cr[i][j] - 128.0)) );
            }
        }

        return R;
    }

    /**
     * This method converts a YUV (aka YCbCr) colorspace to a RGB colorspace.
     * This is handy when trying to reconstruct an image in Java from YCbCr transmitted
     * data. This routine expects the data to fall in the standard PC 0..255 range
     * per pixel, with the array dimensions corresponding to the imageWidth and imageHeight.
     * These variables are either set manually in the case of a null constructor,
     * or they are automatically calculated from the image parameter constructor.
     * @param Y The Y component set.
     * @param Cb The Cb component set.
     * @param Cr The Cr component set.
     * @returns G The G component.
     */
    public int[][] convertYCbCrtoG(int[][] Y, int[][] Cb, int[][] Cr)
    {
        int i;
        int j;
        int G[][] = new int[imageWidth][imageHeight];
        System.out.println(" YCbCr to G");
        for (i=0; i<imageWidth; i++)
        {
            for (j=0; j<imageHeight; j++)
            {
                G[i][j] = (int)(((1.164*((float)Y[i][j]-16.0))) - (0.813*((float)Cr[i][j] - 128.0)) - (0.392*((float)Cb[i][j] - 128.0)));
            }
        }

        return G;
    }

    /**
     * This method converts a YUV (aka YCbCr) colorspace to a RGB colorspace.
     * This is handy when trying to reconstruct an image in Java from YCbCr transmitted
     * data. This routine expects the data to fall in the standard PC 0..255 range
     * per pixel, with the array dimensions corresponding to the imageWidth and imageHeight.
     * These variables are either set manually in the case of a null constructor,
     * or they are automatically calculated from the image parameter constructor.
     * @param Y The Y component set.
     * @param Cb The Cb component set.
     * @param Cr The Cr component set.
     * @returns B The B component.
     */
    public int[][] convertYCbCrtoB(int[][] Y, int[][] Cb, int[][] Cr)
    {
        int i;
        int j;
        int B[][] = new int[imageWidth][imageHeight];
        System.out.println(" YCbCr to B");
        for (i=0; i<imageWidth; i++)
        {
            for (j=0; j<imageHeight; j++)
            {
                B[i][j] = (int)(((1.164*((float)Y[i][j]-16.0))) + (2.017*((float)Cb[i][j] - 128.0)));
            }
        }

        return B;
    }


    /**
     * This method converts a 0..255 range pixel matrix to a java-compatible
     * array of 32 bit integers. This method can take a long time to complete
     * so a reporting feature has been added. The arrays should be the same
     * dimensions as imageWidth and imageHeight. This array can then be converted
     * to an image with: <br>
     * newimage = comp.createImage(new MemoryImageSource(imageWidth,imageHeight,imageArray,0,imageWidth));
     * @param R The Red Values.
     * @param G The Green Values.
     * @param B The Blue Values.
     * @param log Enable/disable progress reporting.
     * @returns array A one dimensional image array.
     */
    public int[] convertRGBtoArray(int[][] R, int[][] G, int[][] B)
    {
        int x = 0;
        int y = 0;
        int index = 0;

        int array[] = new int[imageWidth * imageHeight];

    	for (y=0; y<imageHeight; y++)
    	{
    	    for (x=0; x<imageWidth; x++)
    	    {

                array[index]=(255<<24)|(R[x][y]<<16)|(G[x][y]<<8)|B[x][y];
        	    index++;
    	    }
    	}
        return array;
    }

    /**
     * This method converts a single array 0..255 range pixel matrix to a
     * java-compatible array of 32 bit integers. This method can take a long
     * time to complete so a reporting feature has been added. The arrays
     * should be the same dimensions as imageWidth and imageHeight. This array
     * can then be converted to an image with: <br>
     * newimage = comp.createImage(new MemoryImageSource(imageWidth,imageHeight,imageArray,0,imageWidth));
     * @param Array The pixel values
     * @returns array A one dimensional image array.
     */
    public int[] convertGrayToArray(int[][] R)
    {
        int array[] = new int[imageWidth * imageHeight];

        int x = 0;
        int y = 0;
        int index = 0;

        int G[][] = new int[imageWidth][imageHeight];
        int B[][] = new int[imageWidth][imageHeight];

        G = R;
        B = R;

    	for (y=0; y<imageHeight; y++)
    	{
    	    for (x=0; x<imageWidth; x++)
    	    {
                array[index] = -16777216 | (R[x][y] << 16) | (G[x][y] << 8) | B[x][y];


        	    index++;
    	    }
    	}
        return array;
    }

    public byte[] convertGrayToArray1(int[][] R)
    {
        byte array[] = new byte[imageWidth * imageHeight];

        int x = 0;
        int y = 0;
        int index = 0;

        int G[][] = new int[imageWidth][imageHeight];
        int B[][] = new int[imageWidth][imageHeight];

        G = R;
        B = R;

    	for (y=0; y<imageHeight; y++)
    	{
    	    for (x=0; x<imageWidth; x++)
    	    {
                //array[index] = (byte)-256 | B[x][y];
        	    index++;
    	    }
    	}
        return array;
    }
}


